#include<stdio.h>
#include<stdlib.h>
#include<math.h>
#include<unistd.h>
#include<fcntl.h>

char *ReadLine(int infile) {
  static char buf[1000];
  char t;
  int pos;

  pos=0;
  buf[pos]=0;

  t=0x01;
  while((t!=0x00)&&(t!=0x0A)&&(t!=0x0D)) {
    if (read(infile,&t,1)<=0)
      t=0;
    else
      buf[pos++]=t;
  }
  if (t) {
    pos--;
  }
  buf[pos]=0;
  if (pos)
    return buf;
  else 
    return NULL;
}


int main(int argc,char **argv) {
  double BinSize=1;
  double Min,Max;
  double Val;
  int GotMinMax;
  unsigned long *Bin,BinLen; 
  char *Buf;
  int infile;
  int Lines,Lines2;
  int i,j;
  double Sum,SumSquared;
  double FirstBinVal;
  int AddVal;
  int FileArg=-1;
  int PrintHelp=0;
  int UseValue=0;
  int Graph=0;
  char *p;

  for (i=1;i<argc;i++) {
    switch(argv[i][0]) {
      case '-':
      case '/':
      case '\\':
        switch(argv[i][1]) {
          case 'g': 
          case 'G':
            Graph=1;
            break;
          case 'v':
          case 'V':
            sscanf(&(argv[i][2]),"%i",&UseValue);
            break;
          case 'b':
          case 'B':
            sscanf(&(argv[i][2]),"%lf",&BinSize);
            break;
          case 'h':
          case 'H':
          case '?':
          default:
            PrintHelp=1;
            break;
        }
        break;
      default:
        /* assume this is our filename */
        FileArg=i;
        break;
    }
  }


  if (PrintHelp) {
    printf("USEAGE: %s <filename> [-h] [-b<binsize>] [-v<fieldnum>]\n",argv[0]);
    printf("  This utility will parse a text file, expecting a\n");
    printf("  floating point numerical value in the first column.\n");
    printf("  The output of this program will be the number of lines\n");
    printf("  which fall into each value range (bin). This output is\n");
    printf("  then suitable for graphing. \n");
    printf("   -h   print this help\n");
    printf("   -b   set the bin size\n");
    printf("   -v   turn on value reading. If not used, each line will be\n");
    printf("        counted once. If used, each line's <fieldnum> field is assumed\n");
    printf("        to contain an integer representing the number of times the\n");
    printf("        value should be counted.\n");
    printf("   -g   print a graph instead of values for each bin\n");
    exit(0);
  }

  if (FileArg>=0) {
    infile=open(argv[FileArg],O_RDONLY);
    if (infile<0) {
      printf("%s: cannot open file %s\n",argv[0],argv[1]);
      exit(0);
    }
  } else {
    infile=STDIN_FILENO;
  }

  Max=Min=0;
  GotMinMax=0;
  Lines=0;
  Sum=0.0;
  SumSquared=0.0;

  do {
    Buf=ReadLine(infile);
    if (Buf) {
      Lines++;
      sscanf(Buf,"%lf",&Val);
      if (!GotMinMax) {
        Min=Max=Val;
        GotMinMax=1;
      } else {
        if (Val<Min)
          Min=Val;
        if (Val>Max)
          Max=Val;
      }
    }
  } while (Buf);

  FirstBinVal=((int)(Min/BinSize))*BinSize;
  BinLen=(Max-FirstBinVal)/BinSize+100;
  Bin=malloc(BinLen*sizeof(unsigned long));

  if (!Bin) {
    printf("Cannot allocate enough memory to perform operation.\n");
    exit(0);
  }
  for (i=0;i<BinLen;i++) 
    Bin[i]=0;

  lseek(infile,0,SEEK_SET);


  AddVal=1;
  do {
    Buf=ReadLine(infile);
    if (Buf) {
      if (!UseValue) {
        sscanf(Buf,"%lf",&Val);
        Bin[((int)((Val-FirstBinVal+(BinSize/10))/BinSize))]++;
        Sum+=Val;
        SumSquared+=(Val*Val);
      } else {
        p=Buf;
        for (i=1;i<UseValue;i++) {
          /* find non-whitespace */
          while((*p)&&((*p==' ')||(*p==0x08)))
            p++;
          /*printf("%s\n",p);*/
          /* find end of non-whitespace */
          while((*p)&&(*p!=' ')&&(*p!=0x08))
            p++;
        }
        sscanf(Buf,"%lf",&Val);
        if (!(*p)) {
          printf("Column length exceeded - you might want to check your -v<value> parameter\n");
          AddVal=1;
        } else {
          sscanf(p,"%i",&AddVal);
        }
        Bin[((int)((Val-FirstBinVal+(BinSize/1000))/BinSize))]+=AddVal;
        Sum+=Val*AddVal;
        SumSquared+=(Val*Val*AddVal*AddVal);
      }
    }
  } while (Buf);

  if (!Graph) {
    for (i=((int)((Min-FirstBinVal)/BinSize));i<=((int)((Max-FirstBinVal)/BinSize));i++) {
      printf("%f %li\n",i*BinSize+FirstBinVal,Bin[i]);
    }
  } else {
    Lines2=0;
    for (i=((int)((Min-FirstBinVal)/BinSize));i<=((int)((Max-FirstBinVal)/BinSize));i++) {
      printf("%f :",i*BinSize+FirstBinVal);
      for (j=0;j<Bin[i];j++) 
        printf("*");
      printf("\n");
      Lines2+=Bin[i];
    }
    if ((Lines!=Lines2)&&(!UseValue)) {
      printf("ERROR ERROR ERROR! Lines don't match up. Calculation error. %i %i\n",Lines,Lines2);
    }
  }
  printf("Stepsize=%f, %i lines, Min=%f Max=%f\n",BinSize,Lines,Min,Max);
  printf("Mean=%f StdDev=%f\n",Sum/Lines,sqrt((SumSquared-(Sum*Sum/Lines))/Lines));
  printf("# Bins = %li\n",BinLen-100+1);
  free(Bin);
  close(infile);
  return 0;
}

