#include<stdio.h>

#define IND(x,y)     (x+1+((y+1)*10))
#define INDM(d,x,y)  (x+1+((y+1)*100)+((d+1)*10000))

char *Shape[]={"A","B","C","D","E","F",NULL};
char *Source[]={"one","all",NULL};
char *Traffic[]={"Constant Stream","Event","MultiCon",NULL};

int MakeNet(int ShapeNum,int SourceNum,int TrafficNum, 
  int SizeX,int SizeY,int StopTime, int PacketPeriod,
  int PacketVariance) {
  FILE *outfile;
  char  buf[200];
  char  buf2[200];
  int x,y;

  /* Open file and write our header */
  strcpy(buf2,Traffic[TrafficNum]);
  buf2[4]=0;
  sprintf(buf,"%s_%s_%s.net",Shape[ShapeNum],
    Source[SourceNum],buf2);
  outfile=fopen(buf,"wb");
  if (outfile==NULL) {
    printf("ERROR: Could not open file '%s' for writing!!\n",buf);
    return 1;
  }

  fprintf(outfile,"# THIS .NET FILE CREATED BY NETMAKER\n");
  fprintf(outfile,"# Parameters used to create this file:\n");

  fprintf(outfile,"#   Source:              %s\n",Source[SourceNum]);
  fprintf(outfile,"#   Traffic:             %s\n",Traffic[TrafficNum]);
  fprintf(outfile,"#   StopTime (s):        %i\n",StopTime);
  fprintf(outfile,"#   Packet Period (ms):  %i\n\n",PacketPeriod);
  fprintf(outfile,"#   Network:             %s\n",Shape[ShapeNum]);
  fprintf(outfile,"\n\n");
  fprintf(outfile,"NETWORK\n");
  fprintf(outfile,"  STOPTIME  % 3i\n\n",StopTime);

  /* NodeCaps */
  fprintf(outfile,"  # Let's make some NodeCaps\n");
  for (y=0;y<SizeY;y++) {
    for (x=0;x<SizeX;x++) {
      fprintf(outfile,"  NODECAP   % 4i\n",IND(x,y));
      fprintf(outfile,"    POS       %i %i\n",x+1,y+1);
      if (   ((SourceNum==0)&& (y==0)&&(x==0))  ||
             ((SourceNum==1)&& ((y!=SizeY-1)||(x!=SizeX-1)))
	 ) {
        fprintf(outfile,"    PACKET   \"%s\"\n",Traffic[TrafficNum]);
	switch(TrafficNum) {
	  case 0: /* Constant Stream */
	    fprintf(outfile,"    P0        %i\n",IND(SizeX-1,SizeY-1));
	    fprintf(outfile,"    P1        %i\n",PacketPeriod);
	    break;
	  case 1: /* Event */
	    fprintf(outfile,"    P0        %i\n",IND(SizeX-1,SizeY-1));
	    fprintf(outfile,"    P1        %i\n",PacketPeriod);
	    fprintf(outfile,"    P2        %i\n",PacketVariance);
	    break;
	  case 2: /* MultiCon */
	  default:
	    break;
	}
      }
    }
  }

  /* put in our subnetwork */
  fprintf(outfile,"  # put in our SubNet\n");
  strcpy(buf2,Traffic[TrafficNum]);
  buf2[4]=0;
  sprintf(buf,"%s_%s_%s",Shape[ShapeNum],
    Source[SourceNum],buf2);
  fprintf(outfile,"  SUBNETWORK \"%s\"\n",buf);

  /* define our mediums */
  for (y=0;y<SizeY;y++) {
    for (x=0;x<SizeX;x++) {
      /* put in mediums */
      if (x<(SizeX-1)) {
        fprintf(outfile,"    MEDIUM       %i\n",INDM(0,x,y));
        fprintf(outfile,"      MNAME       \"Wire-%i\"\n",INDM(0,x,y));
        fprintf(outfile,"      TRANSPORT   \"Wire\"\n");
        fprintf(outfile,"      BAUD         100000\n");
        fprintf(outfile,"      DELAY        0.001\n");
      }
      if (y<(SizeY-1)) {
        fprintf(outfile,"    MEDIUM       %i\n",INDM(1,x,y));
        fprintf(outfile,"      MNAME       \"Wire-%i\"\n",INDM(1,x,y));
        fprintf(outfile,"      TRANSPORT   \"Wire\"\n");
        fprintf(outfile,"      BAUD         100000\n");
        fprintf(outfile,"      DELAY        0.001\n");
      }
    }
  }

  /* finally, define our nodes and connect the mediums up */
  fprintf(outfile,"\n\n  # Node Definitions (and network connections) \n");
  for (y=0;y<SizeY;y++) {
    for (x=0;x<SizeX;x++) {
      fprintf(outfile,"    NODE          %i\n",IND(x,y));
      fprintf(outfile,"      PROTOCOL     \"DSR\"\n");
      /* left interface */
      if (x>0)
        fprintf(outfile,"      INTERFACE     %i\n",INDM(0,x-1,y));
      /* top interface */
      if (y>0)
        fprintf(outfile,"      INTERFACE     %i\n",INDM(1,x,y-1));
      /* right interface */
      if (x<(SizeX-1))
        fprintf(outfile,"      INTERFACE     %i\n",INDM(0,x,y));
      /* bottom interface */
      if (y<(SizeY-1))
        fprintf(outfile,"      INTERFACE     %i\n",INDM(1,x,y));
    }
  }
 
  /* now let's just tack on some dummy events and some
   * event help for convenience */
  fprintf(outfile,"\n\n");
  fprintf(outfile,"    # Define our events for this subnetwork\n");
  fprintf(outfile,"    # The format is:\n");
  fprintf(outfile,"    #   EVENT <Event Type> [<Parameter> <Parameter> ...]\n");
  fprintf(outfile,"    # Event types currently defined:\n");
  fprintf(outfile,"    #   EVENT MEDIUM <MediumID> <time> <repeat> <medium state>\n");
  fprintf(outfile,"    #    - bring up or down a link (medium) (ie: enable/disable a link for transmission)\n");
  fprintf(outfile,"    #      <MediumID> = medium ID (which you should have defined above\n");
  fprintf(outfile,"    #      <time> = time of event, in ms\n");
  fprintf(outfile,"    #      <repeat> = if non-zero, the event will be re-scheduled every <repeat> ms\n");
  fprintf(outfile,"    #      <medium state> = defines the medium state: 0=down, 1=up\n");
  fprintf(outfile,"#    EVENT  MEDIUM 10101      1 0 0\n");
  fprintf(outfile,"#    EVENT  MEDIUM 10101   1000 0 1\n\n");
  fprintf(outfile,"    # For Link/Node Addition\n");
  fprintf(outfile,"#    EVENT  MEDIUM 10101      1 0 0\n");
  fprintf(outfile,"#    EVENT  MEDIUM 10101 %i 0 1\n",StopTime*500);
  if (SizeY>1) {
    fprintf(outfile,"#    EVENT  MEDIUM 20101      1 0 0\n");
    fprintf(outfile,"#    EVENT  MEDIUM 20101 %i 0 1\n",StopTime*500);
  }
  fprintf(outfile,"#    EVENT  CONVERGENCE  1 0 0\n");
  fprintf(outfile,"#    EVENT  CONVERGENCE  %i 0 1\n\n",StopTime*500);
  fprintf(outfile,"    # For Link/Node Failure \n");
  fprintf(outfile,"#    EVENT  MEDIUM 2%02i%02i      1 0 0\n",SizeX-1,SizeY);
  fprintf(outfile,"#    EVENT  MEDIUM 2%02i%02i %i 0 1\n",SizeX-1,SizeY,StopTime*500);
  fprintf(outfile,"#    EVENT  MEDIUM 1%02i%02i %i 0 0\n",SizeX,SizeY-1,StopTime*500);
  fprintf(outfile,"#    EVENT  CONVERGENCE  1 0 0\n");
  fprintf(outfile,"#    EVENT  CONVERGENCE  %i 0 1\n\n",StopTime*500);
  
  fprintf(outfile,"\n\n");

  fclose(outfile);
  return 0;
}


int main(int argc,char **argv) {
  int SourceNum;
  int TrafficNum;
  int i;
  int StopTime,PacketPeriod,PacketVariance;
  int Default=0;

  /* parse command-line options */
  for (i=1;i<argc;i++) {
    switch(argv[i][0]) {
      case 'h':
      case 'H':
        printf("USEAGE: %s [<dh>]\n",argv[0]);
	printf("  d  use default network parameters\n");
	printf("  h  this help\n");
	break;
      case 'd':
      case 'D':
        Default=1;
	break;
      default:
        printf("Invalid command line option: %s\n",argv[i]);
	break;
    }
  }

  if (Default) {
    SourceNum=1;
    TrafficNum=0;
    StopTime=60;
    PacketPeriod=900;
    PacketVariance=0;
  } else {
    printf("Source: ");
    for (i=0;Source[i];i++)
      printf("%i)%s ",i,Source[i]);
    printf("  >");
    scanf("%i",&SourceNum);
    if ((SourceNum<0)||(SourceNum>=i)) {
      printf("Invalid Source.\n");
      exit(0);
    }

    printf("Traffic Type: ");
    for (i=0;Traffic[i];i++)
      printf("%i)%s ",i,Traffic[i]);
    printf("  >");
    scanf("%i",&TrafficNum);
    if ((TrafficNum<0)||(TrafficNum>=i)) {
      printf("Invalid Traffic.\n");
      exit(0);
    }
  
    printf("Stop Time (s): ");
    scanf("%i",&StopTime);
    printf("Packet Period (ms): ");
    scanf("%i",&PacketPeriod);
    if (TrafficNum==1) {
      printf("Packet Variance (ms): ");
      scanf("%i",&PacketVariance);
    } else {
      PacketVariance=0;
    }
  }

  MakeNet(0,SourceNum,TrafficNum,2,2,StopTime,PacketPeriod,PacketVariance); 
  MakeNet(1,SourceNum,TrafficNum,4,1,StopTime,PacketPeriod,PacketVariance); 
  MakeNet(2,SourceNum,TrafficNum,3,3,StopTime,PacketPeriod,PacketVariance); 
  MakeNet(3,SourceNum,TrafficNum,9,1,StopTime,PacketPeriod,PacketVariance); 
  MakeNet(4,SourceNum,TrafficNum,4,4,StopTime,PacketPeriod,PacketVariance); 
  MakeNet(5,SourceNum,TrafficNum,16,1,StopTime,PacketPeriod,PacketVariance); 

  printf("Done!\n");
  return 0;
}
