/* This module implements a magic broadcast medium where every packet
 * is broadcast to every node in the network, instantly without
 * collisions or errors. */

/* Options */
/* #define MADEBUG */

#ifdef MADEBUG
#include<stdio.h>
#endif
#include<stdlib.h>
#include"memmgr.h"
#include"network.h"
#include"magicmedium.h"

void maMagicInitMedium(NWMEDIUM *p_Medium) {
  NWINTERFACE *l_Int;
  MAMAGICINTERFACE *l_MagicInt;

#ifdef MADEBUG
  printf("Magic Medium: Init Medium %li\n",p_Medium->mID);
#endif

  for (l_Int=p_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
    l_MagicInt=mmuAlloc(sizeof(MAMAGICINTERFACE));
    l_Int->iGeneral=(void*)l_MagicInt;
    l_MagicInt->mPacket=NULL;
  }
}

void maMagicTickMedium(NWMEDIUM *p_Medium) {
  NWINTERFACE *l_Int,*l_Int2;
  MAMAGICINTERFACE *l_MagicInt;
  NWPACKET *l_Pk;
  long l_Delivered;

#ifdef MADEBUG
  printf("Magic Medium: Tick Medium %li\n",p_Medium->mID);
#endif

  for (l_Int=p_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
    l_MagicInt=l_Int->iGeneral;
    if (l_MagicInt) {
      if (l_MagicInt->mPacket) {
        /* deliver packet */
        l_Delivered=0;
        for (l_Int2=p_Medium->mInterface;l_Int2;l_Int2=l_Int2->iMediumNext) {
          if (l_Int==l_Int2)
            continue;
          if ((l_Int2->iID==l_MagicInt->mPacket->pDestIntID)||
              (l_MagicInt->mPacket->pDestIntID==NWADDR_BROADCAST)) {
            (l_Int2->iNode->nRcvPacket)(l_Int2,l_MagicInt->mPacket);
            l_Delivered++;
          }
        }
        l_Pk=l_MagicInt->mPacket;
        l_MagicInt->mPacket=NULL;
        if (!l_Delivered) {
          /* report error */
          if (l_Int->iNode->nDeliveryError)
            (l_Int->iNode->nDeliveryError)(l_Int,l_Pk,NWDELERR_NOACK);
        }
        nwFreePacket(l_Pk);
      }
    }
  }
}

void maMagicKillMedium(NWMEDIUM *p_Medium) {
  NWINTERFACE *l_Int;

#ifdef MADEBUG
  printf("Magic Medium: Kill Medium %li\n",p_Medium->mID);
#endif

  for (l_Int=p_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
    mmuFree(l_Int->iGeneral);
  }
}

void maMagicKillInterface(NWINTERFACE *p_Interface) {

#ifdef MADEBUG
  printf("Magic Medium: Kill Interface for node %li, medium %li\n",
    p_Interface->iNode->nNodeCap->nID,
    p_Interface->iMedium->mID);
#endif

  mmuFree(p_Interface->iGeneral);
}

unsigned long maMagicSendPacket(NWINTERFACE *p_Interface, NWPACKET *p_Packet, long p_DestinationID) {
  MAMAGICINTERFACE *l_MagicInt;

#ifdef MADEBUG
  printf("Magic Medium: Send Packet from node %li to node %li, via medium %li\n",
    p_Interface->iNode->nNodeCap->nID,
    p_DestinationID,
    p_Interface->iMedium->mID);
#endif

  if (p_Packet) {
    l_MagicInt=p_Interface->iGeneral;
    l_MagicInt->mPacket=nwCopyPacket(p_Packet,1);
    l_MagicInt->mPacket->pSourceIntID=p_Interface->iID;
    l_MagicInt->mPacket->pDestIntID=p_DestinationID;
  }
  return 1;
}
