/* Welcome to my bizarre and twisted network simulator. I made a few
 * mistakes when making this thing, unnecessary complications for the most
 * part, but hey, it was my first wack at writing a simulator. 
 *
 * The way this program organizes a network is as so:
 *  NWNETWORK
 *      |
 *      +-----------NWNODECAP---NWNODECAP---NWNODECAP--...
 *      |               |           |           |
 *   NWSUBNETWORK       |           |           |
 *          |           |           |           |
 *          +---------NWNODE------NWNODE------NWNODE----...
 *                      |           |           |
 *                 NWINTERFACE  NWINTERFACE  NWINTERFACE
 *                      |           |           |
 *                      +--------NWMEDIUM-------+
 * 
 * Now the above is a bit of a simplification of things. 
 * Obviously, any NWNODE can have more than one NWINTERFACE, although
 * at the time of this writing not all modules supported multiple
 * interfaces. 
 * Furthermore, you can have multiple NWSUMBETWORKs for any given
 * NWNETOWRK and group of NWNODECAPs. Why would you want this?
 * Well, Notice within a NWNETOWORK, the NWNODECAPs are constant. 
 * Node locations (for wireless), packet generation, and such are
 * attached to the NWNODECAP. Things like interfaces and mediums
 * (which define the network topology), and network-layer protcols, 
 * are bound to the NWSUBNETWORK. Consequently, you can set up a
 * simulation with multiple NWSUBNETWORKs (which may have different
 * network topologies or protocols) and run them simultaneously. 
 * You can then compare the results quantitatively because the
 * NWSUBNETWORKS are independently operated, but they have the
 * EXACT SAME packet inputs and operating conditions. Cool, eh?
 *
 * To start to understand how this all fits together, look at
 * the header files (in particular, network.h and hwmgr.h) so
 * you can see how the structures reference each other. And then
 * look at the MagicMedium and MagicNet modules - these are
 * "ideal" protocol and physical layers.
 */



#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include<time.h>
#include"network.h"
#include"nsfile.h"
#include"schedule.h"
#include"nam.h"
#include"wire.h"
#include"dsr.h"

/* Prototypes */
void maPrintStat(NWNETWORK *p_Net, long p_Percent);


void mainTimeStampEvent(void* p_Parameter) {
  long t;
  t=((NWNETWORK*)(p_Parameter))->nTick;
  printf("TIMESTAMP: %li tick (time = %li seconds)\n",t,t/NWTICK);
}

int main(int argc, char **argv) {
  NWNETWORK *l_Net;       /* our network */
  NWSUBNETWORK *l_SubNet;
  NWNODECAP *l_NodeCap;
  long l_i,l_PercentDone;
  char l_InFileName[256];
  time_t l_startTime,l_endTime;

  printf("NetSim started.\n");

  /* defaults & Init */
  strcpy(l_InFileName,"netsim.net");
  srand(time(NULL));

  /* read in options */
  for (l_i=1;l_i<argc;l_i++) {
    switch (argv[l_i][0]) {
      case '-':
        printf("This program doesn't take options \n");
        break;
      default:
        /* assume this is an infile */
        strcpy(l_InFileName,argv[l_i]);
        break;
    }
  }

  printf("Reading network file \"%s\":\n",l_InFileName);
  l_Net=nwReadNetFile(l_InFileName);

  /* OK!!!*/

  /* let's initialize our networks */
  nwInitNetwork(l_Net);

  /* And we're done. Hmm.. Now let's just operate the network */
  l_PercentDone=0;
  if (!l_Net) {
    printf("No network to simulate. Exiting.\n");
    exit(0);
  }
  printf("Network:\n");
  for (l_i=0,l_SubNet=l_Net->nSubNetwork;l_SubNet;l_SubNet=l_SubNet->sNext,l_i++);
  printf("   %li subnet%c\n",l_i,l_i==1?' ':'s');
  for (l_SubNet=l_Net->nSubNetwork;l_SubNet;l_SubNet=l_SubNet->sNext) {
    printf("       \"%s\"\n",l_SubNet->sName);
  }
  for (l_i=0,l_NodeCap=l_Net->nNodeCap;l_NodeCap;l_NodeCap=l_NodeCap->nNext,l_i++);
  printf("   %li node%c\n",l_i,l_i==1?' ':'s');
  for (l_NodeCap=l_Net->nNodeCap;l_NodeCap;l_NodeCap=l_NodeCap->nNext) {
    printf("       %li (%li,%li)\n",l_NodeCap->nID,l_NodeCap->nX,
      l_NodeCap->nY);
  }
  printf("   %li seconds, %li ticks (%i tick/s)\n",l_Net->nStopTime/NWTICK,l_Net->nStopTime,NWTICK);
  /*
  printf("   %li pause time (depreciated - don't use in new modules)\n",l_Net->nPauseTime);
  */
  l_startTime=time(NULL);

  /* schedule any permanent events */
  /* This event gives you a regular timestamp message */
  /* schScheduleEvent(NW_S(5),NW_S(5),&mainTimeStampEvent,l_Net);*/

  /* leave this out if you don't want .nam files created (you don't need to 
   * touch the namShutdownNetwork regardless) */
  namInitNetwork(l_Net);

  while(l_Net->nTick<l_Net->nStopTime) {
    nwTickNetwork(l_Net);
    if (!(l_Net->nTick%NW_MS(200))) {
      for (l_SubNet=l_Net->nSubNetwork;l_SubNet;l_SubNet=l_SubNet->sNext) {
        dsrStatPrint(l_SubNet,NWSPFLAG_PRINT | NWSPFLAG_CLEAR);
        wmWireStatPrint(l_SubNet,NWSPFLAG_PRINT | NWSPFLAG_CLEAR);
      }
    }
  }
  namShutdownNetwork(l_Net);

  /* print out stats */
  for (l_SubNet=l_Net->nSubNetwork;l_SubNet;l_SubNet=l_SubNet->sNext) {
    printf("\nSTATS SUBNETWORK %s  ****************************************************\n",l_SubNet->sName);
    dsrStatPrint(l_SubNet,NWSPFLAG_PRINT);
    wmWireStatPrint(l_SubNet,NWSPFLAG_PRINT);
  }


  l_endTime=time(NULL);
  printf("\nFinished. Computed %li simulated seconds in %li real seconds (%4.2f s/r)\n",
    l_Net->nStopTime/NWTICK,l_endTime-l_startTime,
    ((float)(l_Net->nStopTime/NWTICK))/((float)(l_endTime-l_startTime)));
  return 0;
}

