/* memmgr.c
 * ID: mmu
 * The universal memory manager...
 * because nobody seems able to do it right.
 * Copyright (c) 1998 B. Cameron Lesiuk
 */
#include<stdio.h>

#ifdef WIN32
#include<windows.h>
#else
#include<stdlib.h>
#endif
/*#include"memmgr.h"*/

#define MMU_GRANULARITY    (1<<2)  /* boundaries for MMU units to fall on. */
#define MMU_RFLAG_USED     (1<<0)

/*#define MMU_ROUNDUP(x) \          
  (((x)%MMU_GRANULARITY)? \      
   ((x)+MMU_GRANULARITY-((x)%MMU_GRANULARITY)): \  
   (x))                                         */
#define MMU_ROUNDUP(x) ((((unsigned long)(x))%MMU_GRANULARITY)?(((unsigned long)(x))+MMU_GRANULARITY-(((unsigned long)(x))%MMU_GRANULARITY)):((unsigned long)(x)))

#define MMURecord(x) ((MMURECORD*)(x))
#define MMUBlock(x)  ((MMUBLOCK*)(x))
#define MMUVoid(x)   ((void*)(x))

/* A record is an identifying structure at the start of each 
 * memory location. A location may be allocated or free memory.  */
typedef struct MMURECORD {
  unsigned long     rFlag;    /* record flags                             */
  void             *rMemory;  /* pointer to start of memory               */
  unsigned long     rSize;    /* size of memory block EXCLUDING RECORD    */
  struct MMURECORD *rPrev;    /* pointer to previous record in this block */
  struct MMURECORD *rNext;    /* pointer to next record in this block     */
  unsigned long     rTag;     /* for sanity check                         */
} MMURECORD;


/* A block is a chunk of memory allocated from the system. The first thing
 * in this chunk is a MMUBLOCK structure which tells us where to start, 
 * and where the previous and next BLOCKS are (if they exist) */
typedef struct MMUBLOCK {
#ifdef WIN32
  HGLOBAL           bMemoryGlobal;   /* global handle for winshit      */
#endif 
  MMURECORD        *bRecord;   /* pointer to first record              */
  unsigned long     bSize;     /* size of memory block EXCLUDING BLOCK */
  struct MMUBLOCK  *bPrev;     /* pointer to previous block            */
  struct MMUBLOCK  *bNext;     /* Pointer to next block                */
} MMUBLOCK;

/* Globals for this module */
MMUBLOCK      *g_mmuMasterBlock=NULL;      /* start of our memory allocation */
unsigned long  g_mmuBlockSize=(1<<16);     /* master block alloc size        */

/* local prototypes */
void mmuGetRecord(void *p_MemPtr, MMUBLOCK **p_Block, MMURECORD **p_Record);
int mmuVerifyPointer(void *p_MemPtr);

/* This does a sanity check on the MMU data structures */
 void mmuSanityCheck(void) {
  MMUBLOCK       *l_Block;
  MMURECORD      *l_Record;
  static unsigned long l_tagPool=10;
  unsigned long   l_LastFlag,l_NewFlag;

  for (l_Block=g_mmuMasterBlock;
       l_Block;
       l_Block=l_Block->bNext) {
    for (l_Record=l_Block->bRecord;l_Record;l_Record=l_Record->rNext) {
      if (l_Record->rTag==l_tagPool)
        *((int*)(0x00))=1;            
      if (l_Record->rSize>l_Block->bSize)
        *((int*)(0x00))=2;        
      l_Record->rTag=l_tagPool;

      /* check links */
      if (l_Record->rPrev) {
        if (l_Record->rPrev->rNext!=l_Record)
          *((int*)(0x00))=15;        
      }
      if (l_Record->rNext) {
        if (l_Record->rNext->rPrev!=l_Record)
          *((int*)(0x00))=16;        
      }
      /* Check size */
      if (l_Record->rNext)
        if (MMUVoid(l_Record->rMemory)+l_Record->rSize!=l_Record->rNext)
          *((int*)(0x00))=17;        

      /* check also the used flag */
      if (l_Record==l_Block->bRecord) {
        l_LastFlag=l_Record->rFlag&MMU_RFLAG_USED;
      } else {
        l_NewFlag=l_Record->rFlag&MMU_RFLAG_USED;
        if ((!l_LastFlag)&&(!l_NewFlag))
          *((int*)(0x00))=3;        
        l_LastFlag=l_NewFlag;
      }
    }
  }
  l_tagPool++;
}


void *mmuAlloc(unsigned long p_Size) {
  MMUBLOCK       *l_Block;
  MMUBLOCK       *l_LastBlock;
  MMURECORD      *l_Record;
  MMURECORD      *l_NewRecord;
  unsigned long   l_OldBlockSize;
  unsigned long   l_RealBlockSize;
#ifdef WIN32
  HGLOBAL         l_MemoryGlobal;
#endif

  mmuSanityCheck();

  if (p_Size==0)
    return NULL;

  p_Size=MMU_ROUNDUP(p_Size);

  /* first, search for a block with enough space four our alloc */
  for (l_Block=g_mmuMasterBlock;
       l_Block;
       l_Block=l_Block->bNext) {
    /* search records for one that will fit. */
    for (l_Record=l_Block->bRecord;l_Record;l_Record=l_Record->rNext) {
      if ((!(l_Record->rFlag && MMU_RFLAG_USED)) && 
          (l_Record->rSize>=p_Size))
        break;
    }
    if (l_Record)
      break;
  }
  if (!l_Block) { 
    /* we need to alloc a new block!!! */
    /* first, find our last block */
    for (l_LastBlock=NULL,l_Block=g_mmuMasterBlock;
         l_Block;
         l_LastBlock=l_Block,l_Block=l_Block->bNext);
    /* now, l_Block = NULL and l_LastBlock point to the last block in
     * our block chain. Otherwise it's NULL signifying that there are 
     * *no* blocks in our chain!!! */
    /* make sure we get a block big enough for this sucker! */
    l_OldBlockSize=g_mmuBlockSize;
    while(p_Size>(g_mmuBlockSize-MMU_ROUNDUP(sizeof(MMUBLOCK))
      -MMU_ROUNDUP(sizeof(MMURECORD)-MMU_GRANULARITY))) {
      g_mmuBlockSize<<=1;
      if (g_mmuBlockSize<l_OldBlockSize) {
        /* we've wrapped our block size... oops */
        /* the fucker must have asked for like a 4GB page or something. */
        g_mmuBlockSize=l_OldBlockSize;
        mmuSanityCheck();
        return NULL;
      }
    }
    /* alloc it */
#ifdef WIN32
    l_MemoryGlobal=GlobalAlloc(whatever,g_mmuBlockSize);
    if (!l_MemoryGlobal) { 
      g_mmuBlockSize=l_OldBlockSize;
      mmuSanityCheck();
      return NULL;
    }
    l_Block=GlobalLock(l_MemoryGlobal);
    if (!l_Block) {
      GlobalFree(l_MemoryGlobal);
      g_mmuBlockSize=l_OldBlockSize;
      mmuSanityCheck();
      return NULL;
    }
    l_RealBlockSize=some get size function;
#else
    l_Block=malloc(g_mmuBlockSize);
    if (!l_Block) {
      /* can't alloc memory for this sucker! */
      g_mmuBlockSize=l_OldBlockSize;
      mmuSanityCheck();
      return NULL;
    }
    l_RealBlockSize=g_mmuBlockSize;
#endif
    /* initialize this block */
    l_Block->bRecord=MMURecord(MMU_ROUNDUP(MMUVoid(l_Block)+sizeof(MMUBLOCK)));
    l_Block->bSize=l_RealBlockSize-(MMUVoid(l_Block->bRecord)-MMUVoid(l_Block));
    l_Block->bPrev=l_Block->bNext=NULL;
    l_Record=l_Block->bRecord;
    /* initialize first record */
    l_Record->rFlag=0L;
    l_Record->rMemory=MMUVoid(MMU_ROUNDUP(MMUVoid(l_Record)+sizeof(MMURECORD)));
    l_Record->rSize=l_Block->bSize-(MMUVoid(l_Record->rMemory)-MMUVoid(l_Record));
    l_Record->rPrev=l_Record->rNext=NULL;
    l_Record->rTag=0L;
    /* and now link this new block into our block chain */
    if (l_LastBlock) {
      l_LastBlock->bNext=l_Block;
      l_Block->bPrev=l_LastBlock;
      l_Block->bNext=NULL;
    } else { /* this is our *first* block */
      l_Block->bPrev=l_Block->bNext=NULL;
      g_mmuMasterBlock=l_Block;
    }
  }

  /* Ok, huddle! At this point, both our l_Block and l_Record should
   * point to valid structures. Otherwise, we should have aborted by now. */

  /* What we have to do now is break this l_Record up into two parts: our
   * newly used part and, if there's enough left over, an unused (free) part. */
  /* Note: if there isn't enough left over to make an unused (free) part, then
   * we'll just alloc it with the allocated memory - give them a few extra
   * bytes for what its worth. ;) */
  mmuSanityCheck();

  /* Check if this l_Record is big enough to split... */
  if ((l_Record->rSize)>=MMU_ROUNDUP(sizeof(MMURECORD))+MMU_GRANULARITY+p_Size) {
    /* l_Record is big enough to split. */
    l_NewRecord=MMURecord(l_Record->rMemory+p_Size);
    /* set up our new (empty) record */
    l_NewRecord->rFlag=0L;
    l_NewRecord->rMemory=MMUVoid(l_NewRecord)+MMU_ROUNDUP(sizeof(MMURECORD));
    l_NewRecord->rSize=l_Record->rSize-p_Size-MMU_ROUNDUP(sizeof(MMURECORD));
    l_NewRecord->rTag=0L;
    /* insert our new record into the record list */
    l_NewRecord->rPrev=l_Record;
    l_NewRecord->rNext=l_Record->rNext;
    if (l_Record->rNext)
      l_Record->rNext->rPrev=l_NewRecord;
    l_Record->rNext=l_NewRecord;
    /* and adjust our old record size */
    l_Record->rSize=p_Size;
    /* and return it as our memory! */
    l_Record->rFlag|=MMU_RFLAG_USED;
  } else { 
    /* l_Record is just right... alloc the whole record to the request. */
    l_Record->rFlag|=MMU_RFLAG_USED;
  }
  mmuSanityCheck();
  return l_Record->rMemory;
}

void mmuFree(void *p_MemPtr) {
  MMUBLOCK       *l_Block;
  MMURECORD      *l_Record;
  MMURECORD      *l_PrevRecord;
  MMURECORD      *l_NextRecord;
#ifdef WIN32
  HGLOBAL         l_MemoryGlobal;
#endif

  mmuSanityCheck();

  if (!p_MemPtr)
    return;
  mmuGetRecord(p_MemPtr,&l_Block,&l_Record);

  if (l_Record && l_Block) {
    /* This is it!!! */
    /* First, free the memory */
    l_Record->rFlag|=MMU_RFLAG_USED;
    l_Record->rFlag^=MMU_RFLAG_USED;
    /* Second, see if we can re-combine this record into either of its
     * neighbours to form a larger free memory record */
    /* Check our rPrev first... */
    l_PrevRecord=l_Record->rPrev;
    if (l_PrevRecord) {
      if (mmuVerifyPointer(l_PrevRecord)!=-2)
        *((int*)(0x00))=9;
      if (!(l_PrevRecord->rFlag&MMU_RFLAG_USED)) {
        /* YES! We can recombine this with rPrev */
        /* all we have to do is adjust our link... */
        l_PrevRecord->rNext=l_Record->rNext;
        if (l_Record->rNext)
          l_Record->rNext->rPrev=l_PrevRecord;
        /* and our size. */
        l_PrevRecord->rSize+=l_Record->rSize+MMU_ROUNDUP(sizeof(MMURECORD));
        /* and we're done! */
        l_Record=l_PrevRecord;
        if (mmuVerifyPointer(l_Record)!=-2)
          *((int*)(0x00))=10;
      }
    }
    /* Then check rNext */
    /* ie: do pretty much the same thing again, except with rNext... */
    l_NextRecord=l_Record->rNext;
    if (l_NextRecord) {
      if (mmuVerifyPointer(l_NextRecord)!=-2)
        *((int*)(0x00))=11;
      if (!(l_NextRecord->rFlag&MMU_RFLAG_USED)) {
        /* YES! We can recombine this with rNext */
        /* all we have to do is adjust our link... */
        l_Record->rNext=l_NextRecord->rNext;
        if (l_NextRecord->rNext)
          l_NextRecord->rNext->rPrev=l_Record;
        /* and our size. */
        l_Record->rSize+=l_NextRecord->rSize+MMU_ROUNDUP(sizeof(MMURECORD));
        /* and we're done! */
        if (mmuVerifyPointer(l_Record)!=-2)
          *((int*)(0x00))=12;
      }
    }
    mmuSanityCheck();
    /* The last thing we do here is check if our entire block can be
     * deallocated */
    l_Record=l_Block->bRecord;
    if ((!((l_Record->rFlag)&MMU_RFLAG_USED))&&
        (l_Record->rNext==NULL)) {
      /* This entire block is free - let's free it back to the system */

      /* first, let's remove it from the block link */
      if (l_Block->bPrev) {
        l_Block->bPrev->bNext=l_Block->bNext;
      } else {
        g_mmuMasterBlock=l_Block->bNext;
      }
      if (l_Block->bNext) {
        l_Block->bNext->bPrev=l_Block->bPrev;
      }

      /* second, let's actually free the sucker!!! Yay! */
#ifdef WIN32
      l_MemoryGlobal=l_Block->bMemoryGlobal;
      GlobalLock(l_MemoryGlobal);
      GlobalFree(l_MemoryGlobal);
#else
      free(l_Block);
#endif
    }
  }  else {
    *((int*)(0x00))=8;
  }/* else invalid pointer -- ignore it. */
  mmuSanityCheck();
}

void *mmuReAlloc(void *p_MemPtr, unsigned long p_NewSize) {
  void           *l_OldMem;
  void           *l_NewMem;
  void           *l_NewMem2;
  MMUBLOCK       *l_Block;
  MMURECORD      *l_Record;
  MMURECORD      *l_NewRecord;
  MMURECORD      *l_NextRecord;
  unsigned long   l_i;
#ifdef WIN32
  HGLOBAL         l_MemoryGlobal;
#endif

  mmuSanityCheck();

  if (!p_MemPtr) {
    return mmuAlloc(p_NewSize);
  }
  if (!p_NewSize) {
    /* just do a free */
    mmuFree(p_MemPtr);
    mmuSanityCheck();
    return NULL;
  }

  p_NewSize=MMU_ROUNDUP(p_NewSize);

  mmuGetRecord(p_MemPtr,&l_Block,&l_Record);

  if (l_Record && l_Block) {
    /* ok, we found it. */
    if (l_Record->rSize>p_NewSize) {
      /* let's try to just shrink this record. */
      if ((l_Record->rNext)&&(!(l_Record->rNext->rFlag&MMU_RFLAG_USED))) {
        /* we can recombine the freed space with our next record */
        l_NextRecord=l_Record->rNext;
        l_NewRecord=MMURecord(l_Record->rMemory+p_NewSize);
        l_NewRecord->rFlag=l_NextRecord->rFlag;
        l_NewRecord->rMemory=MMUVoid(l_NewRecord)+
          MMU_ROUNDUP(sizeof(MMURECORD));
	l_NewRecord->rSize=l_NextRecord->rSize+
	  (l_NextRecord->rMemory-l_NewRecord->rMemory);
	l_NewRecord->rTag=0L;
        l_Record->rSize=p_NewSize;
	l_NewRecord->rPrev=l_Record;
	l_NewRecord->rNext=l_NextRecord->rNext;
	l_Record->rNext=l_NewRecord;
	if (l_NewRecord->rNext)
	  l_NewRecord->rNext->rPrev=l_NewRecord;
      } else {
        /* we have to try and create a new record for the extra space */
        if ((l_Record->rSize-MMU_ROUNDUP(p_NewSize))>
          (MMU_ROUNDUP(sizeof(MMURECORD))+MMU_GRANULARITY)) {
          /* Enough room to make a record - let's do so */
          l_NewRecord=MMURecord(l_Record->rMemory+p_NewSize);
          l_NewRecord->rFlag=0L;
          l_NewRecord->rMemory=MMUVoid(l_NewRecord)+
            MMU_ROUNDUP(sizeof(MMURECORD));
          l_NewRecord->rSize=l_Record->rSize-
            (l_NewRecord->rMemory-l_Record->rMemory);
          l_NewRecord->rTag=0L;
          l_NewRecord->rNext=l_Record->rNext;
          l_Record->rNext=l_NewRecord;
          if (l_NewRecord->rNext)
            l_NewRecord->rNext->rPrev=l_NewRecord;
          l_Record->rSize=p_NewSize;
        } /* else not enough room to make a new record - leave it */
      }
      /* either way, we leave the data where it was. */
      mmuSanityCheck();
      return l_Record->rMemory;
    } else if (l_Record->rSize<p_NewSize) {
      /* we need a larger memory location for this sucker. */
      l_NextRecord=l_Record->rNext;
      if ((l_NextRecord)&&
	  ((l_Record->rSize+l_NextRecord->rSize+
	  MMU_ROUNDUP(sizeof(MMURECORD)))>=p_NewSize)) {
	/* expand into next (free) record */
	l_Record->rNext=l_NextRecord->rNext;
	if (l_Record->rNext)
	  l_Record->rNext->rPrev=l_Record;
        l_Record->rSize+=l_NextRecord->rSize+MMU_ROUNDUP(sizeof(MMURECORD));
	/* now see if we can reclaim some of the new record back into free
	 * space. */
	if ((l_Record->rSize-p_NewSize)>= 
	  (MMU_ROUNDUP(sizeof(MMURECORD))+MMU_GRANULARITY)) {
	  /* Yes, we can make a new record from this */
	  l_NewRecord=l_Record->rMemory+p_NewSize;
	  l_NewRecord->rMemory=MMUVoid(l_NewRecord)+
	    MMU_ROUNDUP(sizeof(MMURECORD));
	  l_NewRecord->rSize=l_Record->rSize-p_NewSize-
	    MMU_ROUNDUP(sizeof(MMURECORD));
	  l_Record->rSize=p_NewSize;
	  l_NewRecord->rNext=l_Record->rNext;
	  l_Record->rNext=l_NewRecord;
	  l_NewRecord->rPrev=l_Record;
	  if (l_NewRecord->rNext)
	    l_NewRecord->rNext->rPrev=l_NewRecord;
	}
        mmuSanityCheck();
	return l_Record->rMemory;
      } else {
	/* just alloc a new area, copy it, and return */
        l_NewMem2=l_NewMem=mmuAlloc(p_NewSize);
	if (!l_NewMem) {
          mmuSanityCheck();
	  return NULL; /* couldn't expand - not enough memory */
	}
	/* copy from old to new */
	l_OldMem=l_Record->rMemory;
	for (l_i=0;l_i<l_Record->rSize;l_i++) {
	  *((char *)(l_NewMem))=*((char*)(l_OldMem));
	  l_NewMem++;
	  l_OldMem++;
        }
	/* Free old memory */
	mmuFree(l_Record->rMemory);
	/* return new handle */
        mmuSanityCheck();
	return l_NewMem2;
      }
    } else {
      /* They want it the same size... just return the old pointer. */
      mmuSanityCheck();
      return l_Record->rMemory;
    }
  }
  mmuSanityCheck();
  return NULL;
}

unsigned long mmuGetSize(void *p_MemPtr) {
  MMUBLOCK       *l_Block;
  MMURECORD      *l_Record;

  mmuSanityCheck();
  if (!p_MemPtr)
    return 0L;
  mmuGetRecord(p_MemPtr,&l_Block,&l_Record);
  if (l_Record && l_Block) {
    return (l_Record->rSize);
  }
  return 0L;
}

/* retrieves block and record for a given pointer */
void mmuGetRecord(void *p_MemPtr, MMUBLOCK **p_Block, MMURECORD **p_Record) {
  MMUBLOCK       *l_Block;
  MMURECORD      *l_Record;

  /* first, find the block (if any) to which this pointer belongs */
  for (l_Block=g_mmuMasterBlock;
       l_Block;
       l_Block=l_Block->bNext) {
    if ((p_MemPtr>=MMUVoid(l_Block->bRecord))&&
        (p_MemPtr<(MMUVoid(l_Block->bRecord)+l_Block->bSize))) {
      /* This is our block! */
      break;
    }
  }
  if (l_Block) {
    /* ok, now let's find its record */
    for (l_Record=l_Block->bRecord;
         l_Record;
         l_Record=l_Record->rNext) {
      /*if ((p_MemPtr>=MMUVoid(l_Record->rMemory))&&*/
      if ((p_MemPtr>=MMUVoid(l_Record))&&
          (p_MemPtr<(l_Record->rMemory+l_Record->rSize))) {
        break;
      }
    }
    if (l_Record) { 
      /* This is it! */
      *p_Block=l_Block;
      *p_Record=l_Record;
      return;
    }
  }
  *p_Block=NULL;
  *p_Record=NULL;
}

void mmuPrintData(void) {
  MMUBLOCK  *l_Block;
  MMURECORD *l_Record;
  unsigned long l_tag;

  printf("Memory Allocation:\n");
  for (l_Block=g_mmuMasterBlock;l_Block;l_Block=l_Block->bNext) {
    /* summarize this block */
    printf("BLOCK %p  size: %lX  prev: %p  next: %p\n",
      l_Block,l_Block->bSize,l_Block->bPrev,l_Block->bNext);
    l_tag=0xfffffffe;
    for (l_Record=l_Block->bRecord;l_Record;l_Record=l_Record->rNext) {
      printf("    RECORD %p %s\n     size: %lX",l_Record,
        (l_Record->rFlag&MMU_RFLAG_USED)?"allocated":"free",
        l_Record->rSize);
      printf("  prev: %p  next: %p\n",l_Record->rPrev,l_Record->rNext);
      if (l_Record->rTag==l_tag)
        *((int*)(0x00))=3;
      l_tag=l_Record->rTag=l_Record->rTag-1;
    }
  }
}

int mmuVerifyPointer(void *p_MemPtr) {
  MMUBLOCK *l_Block;
  MMURECORD *l_Record;

  mmuGetRecord(p_MemPtr,&l_Block,&l_Record);

  if (l_Block)  {
    if (p_MemPtr==l_Record->rMemory)
      return 1; /* pointer is perfect! */
    else if (p_MemPtr==l_Record)
      return -2;
    else
      return -1; /* pointer points to someplace in the middle */
  } 
  return 0; /* pointer is completely invalid */
}
