/* nam.c
 * ID: nam
 * This module supports the creation of .nam files to use with the nam 
 * network animation tool. 
 * WARNING NOTE WARNING NOTE : There is a lot of code in here
 *  that only works on point-to-point links (ie links with only 2
 *  interfaces attached). This is only because I don't know how
 *  to get nam to work with multi-drop shared mediums. 
 *  
 */

#include<stdio.h>
#include<sys/types.h>
#include<sys/stat.h>
#include<fcntl.h>
#include<unistd.h>
#include<string.h>
#include"network.h"
#include"schedule.h"
#include"memmgr.h"

char *namColour[]={
  "red",
  "green",
  "blue",
  "black",
  ""
};

/* This function opens up a .nam file for each subnetwork. The name of the .nam file
 * is derived from the subnetwork's sName. Each .nam file is opened and all the
 * header information is written into the file. Refer to the documentation for
 * nam to find out all the nasty details about the .nam file format. 
 */
void namInitNetwork(NWNETWORK *p_Net) {
  NWSUBNETWORK *l_SubNet;
  NWNODECAP *l_NodeCap,*l_NodeCap2;
  NWMEDIUM *l_Medium;
  NWINTERFACE *l_Interface;
  char l_buf[1024],l_buf2[256];
  int l_i,l_j;
  float l_queue=0;
  int l_angle;

  if (!p_Net)
    return;

  for (l_SubNet=p_Net->nSubNetwork;l_SubNet;l_SubNet=l_SubNet->sNext) {
    /* open a file for this subnetwork's .nam file */
    sprintf(l_buf,"%s.nam",l_SubNet->sName);
    unlink(l_buf);
    l_SubNet->sFile=open(l_buf,O_RDWR|O_CREAT|O_TRUNC,S_IRUSR|S_IWUSR|S_IRGRP|S_IROTH);
    if (l_SubNet->sFile>=0) {
      /* Write header info for this .nam file */

      /* Version and Annotation Information (I think) */
      sprintf(l_buf,"V -t * -v 1.0a5 -a 0\n");
      write(l_SubNet->sFile,l_buf,strlen(l_buf));

      /* Why the hell do we need these?! */
      sprintf(l_buf,"A -t * -n 1 -p 0 -o 255 -c 15 -a 1\n");
      write(l_SubNet->sFile,l_buf,strlen(l_buf));
      sprintf(l_buf,"A -t * -h 1 -m 127 -s 8\n");
      write(l_SubNet->sFile,l_buf,strlen(l_buf));

      /* Packet (Attribute) Colours */
      for (l_i=0;namColour[l_i][0];l_i++) {
        sprintf(l_buf,"c -t * -i %i -n %s\n",l_i,namColour[l_i]);
        write(l_SubNet->sFile,l_buf,strlen(l_buf));
      }

      /* Node definitions - shapes can be circle, hexagon, box */
      for (l_NodeCap=p_Net->nNodeCap;l_NodeCap;l_NodeCap=l_NodeCap->nNext) {
        if (l_NodeCap->nTickNodeCap) { /* will be non-null if we have a packet generator */
          sprintf(l_buf,"n -t * -a %li -s %li -S UP -v box -c black\n",l_NodeCap->nID,l_NodeCap->nID);
	} else {
          sprintf(l_buf,"n -t * -a %li -s %li -S UP -v circle -c black\n",l_NodeCap->nID,l_NodeCap->nID);
	}
        write(l_SubNet->sFile,l_buf,strlen(l_buf));
      }

      /* Link definitions - note this will only link first 2 devices hooked to a medium. Sigh. */
      for (l_Medium=l_SubNet->sMedium;l_Medium;l_Medium=l_Medium->mNext) {
        l_Interface=l_Medium->mInterface;
	l_i=l_j=-1;
	l_buf2[0]=0;
	if (l_Interface) {
	  l_NodeCap=l_Interface->iNode->nNodeCap;
	  l_i=l_NodeCap->nID;
	  l_Interface=l_Interface->iMediumNext;
	}
	if (l_Interface) {
	  l_NodeCap2=l_Interface->iNode->nNodeCap;
	  l_j=l_NodeCap2->nID;
	  if (l_NodeCap->nX<l_NodeCap2->nX)  {
	    strcat(l_buf2,"right");
	  } else if (l_NodeCap->nX>l_NodeCap2->nX) {
	    strcat(l_buf2,"left");
	  }
	  if ((l_NodeCap->nY!=l_NodeCap2->nY)&&(l_buf2[0])) {
	    strcat(l_buf2,"-"); 
          }
	  if (l_NodeCap->nY<l_NodeCap2->nY) {
	    strcat(l_buf2,"down");
	  } else if (l_NodeCap->nY>l_NodeCap2->nY) {
	    strcat(l_buf2,"up");
	  }
	}

        sprintf(l_buf,"l -t * -s %i -d %i -S UP -r %f -D %f -c black -o %s\n",
	  l_i,l_j,                        /* source , destination */
	  l_Medium->mBaud,                /* baud rate of link, in bits per second */
	  l_Medium->mPropagationDelay,    /* propagation delay, in seconds. */
	  l_buf2);                        /* layout direction */
        write(l_SubNet->sFile,l_buf,strlen(l_buf));
      }

      /* Queue position definitions */
      for (l_Medium=l_SubNet->sMedium;l_Medium;l_Medium=l_Medium->mNext) {
        l_Interface=l_Medium->mInterface;
	l_i=l_j=-1;
	if (l_Interface) {
	  l_NodeCap=l_Interface->iNode->nNodeCap;
	  l_i=l_NodeCap->nID;
	  l_Interface=l_Interface->iMediumNext;
	}
	if (l_Interface) {
	  l_NodeCap2=l_Interface->iNode->nNodeCap;
	  l_j=l_NodeCap2->nID;
	  l_queue=0.0;
	  l_angle=0;
	  if (l_NodeCap->nX<l_NodeCap2->nX)  {
	    l_queue=0.0;
	    l_angle=1;
	  } else if (l_NodeCap->nX>l_NodeCap2->nX) {
	    l_queue=1.0;
	    l_angle=2;
	  }
	  if (l_NodeCap->nY<l_NodeCap2->nY) {
	    if (l_angle) {
	      l_queue+=0.5;
	      l_queue/=2;
	    } else 
	      l_queue=0.5;
	    l_angle++;
	  } else if (l_NodeCap->nY>l_NodeCap2->nY) {
	    if (l_angle==1) {
	      l_queue=1.75;
	    } else if (l_angle==2) {
	      l_queue=1.25;
	    } else 
	      l_queue=1.5;
	    l_angle++;
	  }
          l_queue+=0.25;
	  if (l_queue>=2.0)
	    l_queue-=2.0;
	}

	sprintf(l_buf,"q -t * -s %i -d %i -a %f\n",l_i,l_j,l_queue);
        write(l_SubNet->sFile,l_buf,strlen(l_buf)); 
	if (l_queue>=1.0)
	  l_queue-=1.0;
	else
	  l_queue+=1.0;
	sprintf(l_buf,"q -t * -s %i -d %i -a %f\n",l_j,l_i,l_queue);
        write(l_SubNet->sFile,l_buf,strlen(l_buf)); 
      }
      /* and that should do us. Now we just need TRAFFIC!! */
    }
  }
}

/* This just closes all the open .nam files before we quit */
void namShutdownNetwork(NWNETWORK *p_Net) {
  NWSUBNETWORK *l_SubNet;
  if (p_Net) {
    for (l_SubNet=p_Net->nSubNetwork;l_SubNet;l_SubNet=l_SubNet->sNext) {
      if (l_SubNet->sFile>=0) {
        close(l_SubNet->sFile);
	l_SubNet->sFile=-1;
      }
    }
  }
}


/* This function writes into the .nam file the appropriate junk to 
 * animate the transmission of p_Packet on p_Medium. The rest of
 * the info we need to make this happen can be extracted from
 * the p_Medium and p_Packet data structures. :) */
void namPacketEvent(NWMEDIUM *p_Medium, NWPACKET *p_Packet,long p_TicksToSend,char p_Event) {
  NWSUBNETWORK *l_SubNet;
  NWINTERFACE *l_Int;
  long l_Src,l_Dst;
  char l_buf[256];
  float l_time;

  if ((!p_Medium)||(!p_Packet))
    return;
  
  l_SubNet=p_Medium->mSubNetwork;
  if (l_SubNet->sFile>=0) {
    l_Src=l_Dst=NWADDR_BROADCAST;
    for (l_Int=p_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
      if ((l_Int->iID==p_Packet->pSourceIntID) &&(l_Src==NWADDR_BROADCAST))
        l_Src=l_Int->iNode->nNodeCap->nID;
      if ((l_Int->iID==p_Packet->pDestIntID)&&(l_Dst==NWADDR_BROADCAST))
        l_Dst=l_Int->iNode->nNodeCap->nID;
    }
    if (l_Dst==NWADDR_BROADCAST) { /* default to first non-source Interface ID */
      for (l_Int=p_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
        if (l_Int->iID!=p_Packet->pSourceIntID) {
          l_Dst=l_Int->iNode->nNodeCap->nID;
	  break;
	}
      }
    }

    l_time=l_SubNet->sNet->nTick;
    l_time/=NWTICK;
    /* write this sucker to our nam file */
    sprintf(l_buf,"%c -t %f -s %li -d %li -p %s -e %li -c %li -i %li -a %li\n",
      p_Event,                       /* our event type */
      l_time,                        /* time */
      l_Src,                         /* source */
      l_Dst,                         /* destination */
      p_Packet->pNamString,          /* traffic source */
      p_Packet->pDataLen+p_Packet->pPayloadLen, /* packet size */
      p_Packet->pNamConversationID,  /* conversation ID */
      p_Packet->pID,                 /* packet ID in the conversation - unique */
      p_Packet->pNamColour);         /* attribute (currently used as color ID) */
    write(l_SubNet->sFile,l_buf,strlen(l_buf)); 
  }
}

void namPacketTransmit(NWMEDIUM *p_Medium, NWPACKET *p_Packet,long p_TicksToSend) {
  namPacketEvent(p_Medium,p_Packet,p_TicksToSend,'h');
}

void namPacketReceive(NWMEDIUM *p_Medium, NWPACKET *p_Packet,long p_TicksToSend) {
  namPacketEvent(p_Medium,p_Packet,p_TicksToSend,'r');
}

void namPacketDiscard(NWMEDIUM *p_Medium, NWPACKET *p_Packet,long p_TicksToSend) {
  namPacketEvent(p_Medium,p_Packet,p_TicksToSend,'d');
}

void namLinkDown(NWMEDIUM *p_Medium) {
  NWSUBNETWORK *l_SubNet;
  NWINTERFACE *l_Int;
  long l_Src,l_Dst;
  char l_buf[256];
  float l_time;

  if (!p_Medium) 
    return;
  l_SubNet=p_Medium->mSubNetwork;
  if (l_SubNet->sFile>=0) {
    l_Src=l_Dst=NWADDR_BROADCAST;
    l_Int=p_Medium->mInterface;
    if (l_Int) {
      l_Src=l_Int->iNode->nNodeCap->nID;
      l_Int=l_Int->iMediumNext;
      if (l_Int)
        l_Dst=l_Int->iNode->nNodeCap->nID;
    }

    l_time=l_SubNet->sNet->nTick;
    l_time/=NWTICK;
    sprintf(l_buf,"l -t %f -s %li -d %li -S DOWN\n",
      l_time,      /* time             */
      l_Src,       /* source node      */
      l_Dst);      /* destination node */
    write(l_SubNet->sFile,l_buf,strlen(l_buf)); 
  }
}

void namLinkUp(NWMEDIUM *p_Medium) {
  NWSUBNETWORK *l_SubNet;
  NWINTERFACE *l_Int;
  long l_Src,l_Dst;
  char l_buf[256];
  float l_time;

  if (!p_Medium) 
    return;
  l_SubNet=p_Medium->mSubNetwork;
  if (l_SubNet->sFile>=0) {
    l_Src=l_Dst=NWADDR_BROADCAST;
    l_Int=p_Medium->mInterface;
    if (l_Int) {
      l_Src=l_Int->iNode->nNodeCap->nID;
      l_Int=l_Int->iMediumNext;
      if (l_Int)
        l_Dst=l_Int->iNode->nNodeCap->nID;
    }

    l_time=l_SubNet->sNet->nTick;
    l_time/=NWTICK;
    sprintf(l_buf,"l -t %f -s %li -d %li -S UP\n",
      l_time,      /* time             */
      l_Src,       /* source node      */
      l_Dst);      /* destination node */
    write(l_SubNet->sFile,l_buf,strlen(l_buf)); 
  }
}


void namQueueEvent(NWNODE *p_Node, NWPACKET *p_Packet,long p_TicksToSend,char p_Event) {
  NWSUBNETWORK *l_SubNet;
  NWINTERFACE *l_Int;
  NWMEDIUM *l_Medium;
  long l_Src,l_Dst;
  char l_buf[256];
  float l_time;

  if ((!p_Node)||(!p_Packet))
    return;
  
  l_SubNet=p_Node->nSubNet;
  if (l_SubNet->sFile>=0) {
    l_Src=p_Node->nNodeCap->nID;
    l_Dst=NWADDR_BROADCAST;
    l_Medium=p_Node->nInterface->iMedium;
    for (l_Int=l_Medium->mInterface;l_Int;l_Int=l_Int->iMediumNext) {
      if (l_Int->iNode->nNodeCap->nID!=l_Src) {
        l_Dst=l_Int->iNode->nNodeCap->nID;
	break;
      }
    }

    l_time=l_SubNet->sNet->nTick;
    l_time/=NWTICK;
    /* write this sucker to our nam file */
    sprintf(l_buf,"%c -t %f -s %li -d %li -p %s -e %li -c %li -i %li -a %li\n",
      p_Event,                       /* our event type */
      l_time,                        /* time */
      l_Src,                         /* source */
      l_Dst,                         /* destination */
      p_Packet->pNamString,          /* traffic source */
      p_Packet->pDataLen+p_Packet->pPayloadLen, /* packet size */
      p_Packet->pNamConversationID,  /* conversation ID */
      p_Packet->pID,                 /* packet ID in the conversation - unique */
      p_Packet->pNamColour);         /* attribute (currently used as color ID) */
    write(l_SubNet->sFile,l_buf,strlen(l_buf)); 
  }
}

void namPacketQueueAdd(NWNODE *p_Node, NWPACKET *p_Packet) {
  namQueueEvent(p_Node,p_Packet,0,'+');
}

void namPacketQueueRemove(NWNODE *p_Node, NWPACKET *p_Packet) {
  namQueueEvent(p_Node,p_Packet,0,'-');
}

void namPacketQueueDiscard(NWNODE *p_Node, NWPACKET *p_Packet) {
  namQueueEvent(p_Node,p_Packet,0,'d');
}


