/* nsfile.c
 * two-letter id: nf
 */
/* NetSim File 
 * This module is designed to read in a .net file and construct a
 * NWNETWORK structure from it. 
 */

#include<stdio.h>
#include<stdlib.h>
#include<string.h>
#include"network.h"
#include"hwmgr.h"
#include"nsfile.h"
#include"schedule.h"

int stricmp(char *l_s1,char *l_s2) {
  char l_t1,l_t2;
  while(1) {
    l_t1=*l_s1;
    l_t2=*l_s2;
    if ((l_t1>='A')&&(l_t1<='Z')) l_t1=l_t1+('a'-'A');
    if ((l_t2>='A')&&(l_t2<='Z')) l_t2=l_t2+('a'-'A');

    if ((l_t1==0)&&(l_t2==0))
      return 0;

    if (l_t1>l_t2)
      return 1;

    if (l_t1<l_t2)
      return -1;

    l_s1++;
    l_s2++;
  }
}

char *nwAdvanceToNonSpace(char *p_Line) {
  char *l_p;

  l_p=p_Line;
  /* advance to the first non-blank character */
  while(((*l_p==0x20)||(*l_p==0x09))&&(*l_p))
    l_p++;

  return l_p;
}

char *nwGetFirstQuote(char *p_Line, char *p_buf) {
  char *l_p,*l_q,*l_b;

  l_p=nwAdvanceToNonSpace(p_Line);

  if (!(*l_p)) {
    p_buf[0]=0;
    return l_p;
  }

  l_q=l_p+1;
  while((*l_q)&&(*l_q!=*l_p)) {
  /*  if ((*l_q>='A')&&(*l_q<='Z'))
    *l_q=*l_q+('a'-'A'); */
    l_q++;
  }
  /* l_q now points to 1 past our last word character */
  /* Copy to p_buf */
  l_p++;
  for (l_b=p_buf;l_p<l_q;l_p++,l_b++) {
    *l_b=*l_p;
  }
  *l_b=0;

  /* finally, advance past last quote mark */
  if (*l_q)
  l_q++;

  return l_q;
}

char *nwGetFirstWord(char *p_Line, char *p_buf) {
  char *l_p,*l_q,*l_b;

  l_p=nwAdvanceToNonSpace(p_Line);

  if (!(*l_p)) {
    p_buf[0]=0;
    return l_p;
  }

  l_q=l_p;
  while((*l_q)&&(*l_q!=0x20)&&(*l_q!=0x09)) {
    if ((*l_q>='A')&&(*l_q<='Z'))
    *l_q=*l_q+('a'-'A');
    l_q++;
  }
  /* l_q now points to 1 past our last word character */
  /* Copy to p_buf and exit */
  l_b=p_buf;
  for (l_b=p_buf;l_p<l_q;l_p++,l_b++) {
    *l_b=*l_p;
  }
  *l_b=0;
  return l_q;
}

NWNETWORK *nwReadNetFile(char *p_File) {
  FILE *l_infile;
  NWNETWORK *l_Net;
  NWSUBNETWORK *l_SubNet;
  NWMEDIUM *l_Medium,*l_m;
  NWINTERFACE *l_Interface;
  NWNODE *l_Node,*l_n;
  NWNODECAP *l_NodeCap,*l_nc;
  char l_File[256];
  char l_buf[256],l_word[256];
  char *l_p;
  long l_i,l_Exp,l_Rep,l_State;
 
  if (!p_File) 
    return NULL;
 
  l_Net=NULL;
  l_SubNet=NULL;
  l_Medium=NULL;
  l_Interface=NULL;
  l_Node=NULL;
  l_NodeCap=NULL;

  strcpy(l_File,p_File);

  l_infile=fopen(l_File,"rb");
  if (!l_infile) {
    strcat(l_File,".net");
    l_infile=fopen(l_File,"rb");
  }  
  if (!l_infile) {
    printf("File named %s or %s cannot be opened.\n",p_File,l_File);
    return NULL;
  }
  /* Ok, let's read this sucker in, shall we? */
  while(!feof(l_infile)) {
    /* advance past carridge returns & line feeds */
    l_buf[0]=0x0D;
    while(((l_buf[0]==0x0A)||(l_buf[0]==0x0D))&&!feof(l_infile)) {
      fread(&(l_buf[0]),1,1,l_infile);
    }
    if (feof(l_infile))
      continue;         /* will exit loop */

    /* read a line */
    l_p=l_buf;
    do {
      l_p++;
      fread(l_p,1,1,l_infile);
    } while((!feof(l_infile))&&(*l_p!=0x0A)&&(*l_p!=0x0D));
    *l_p=0;

    /* process this line */
    /* Get first word */
    l_p=nwGetFirstWord(l_buf,l_word);
    if (!l_word[0])
      continue;
    /* ok! This is our first "meat" information! Process it. */
    switch(l_word[0]) {
      case '#':
	/* line ignored. */
	continue;
      case 'n': /* NETWORK, NODECAP, NODE */
	if (l_word[1]=='e') { /* NETWORK */
	  if (!l_Net) {
	    /* alloc a new network structure! Yay! */
	    l_Net=nwAllocNetwork();
	  } else {
	    printf("ERROR: Multiple NETWORKs specified\n");
	    printf("       Sorry, multiple NETWORKs are unsupported!!\n");
	  }
	  break;
	} else if (l_word[1]=='o') { /* NODECAP, NODE */
	  if (strlen(l_word)<=4) {  /* NODE */
	    if (l_SubNet) {
            /* read in our nodecap ID */
              l_p=nwGetFirstWord(l_p,l_word);
	      l_i=0;
	      l_Node=NULL;
              if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	        /* read in ID */
	        sscanf(l_word,"%li",&l_i);
	        /* now find that node */
	        for (l_n=l_SubNet->sNode;
	          (l_n)&&(l_n->nNodeCap->nID!=l_i);
	          l_n=l_n->nNext);
	        if (l_n) {
	          l_Node=l_n;
	        }
              } 
	    } else {
	      printf("ERROR: NODE used without previous SUBNET\n");
	    }
	  } else {    /* NODECAP */
	    if (l_Net) {
	      /* add a nodecap to this sucker. */
	      l_NodeCap=nwAllocNodeCap(l_Net);
	      /* check for ID specified */
              l_p=nwGetFirstWord(l_p,l_word);
              if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	        /* read in new ID */
	        sscanf(l_word,"%li",&(l_NodeCap->nID));
	      }
	      /* ignore rest of line */
	    } else {
	      printf("ERROR: NODECAP used without previous NETWORK\n");
	    }
	  }
	} else {
	  printf("ERROR: Unknown command %s in file %s.\n",l_word,l_File);
	}
	break;
      case 'p': /* POS, PACKET_ALGORYTHM, PROTOCOL, PAUSETIME, P0-P9 */
        if ((l_word[1]>='0')&&(l_word[1]<='9')) { /* P0 - P9 */
	  if (l_NodeCap) { 
	    l_i=l_word[1]-'0';
	    /* read in the parameter */
            l_p=nwGetFirstWord(l_p,l_word);
            if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	      /* read in parameter value */
	      sscanf(l_word,"%li",&(l_NodeCap->nParam[l_i]));
	    }
	  } else {
	    printf("ERROR: P<#> parameter used without previous NODECAP\n");
	  }
	} else if (l_word[1]=='o') { /* POS */
	  if (l_NodeCap) { 
	    /* read in X and Y coords */
            l_p=nwGetFirstWord(l_p,l_word);
            if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	      /* read in new ID */
	      sscanf(l_word,"%li",&(l_NodeCap->nX));
	      if (l_NodeCap->nX>0)
	        l_NodeCap->nX=NW_MM(l_NodeCap->nX);
	    }
            l_p=nwGetFirstWord(l_p,l_word);
            if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	      /* read in new ID */
	      sscanf(l_word,"%li",&(l_NodeCap->nY));
	      if (l_NodeCap->nY>0)
	        l_NodeCap->nY=NW_MM(l_NodeCap->nY);
	    }
	  } else {
	    printf("ERROR: POS used without previous NODECAP\n");
	  }
	} else if (l_word[1]=='r') { /* PROTOCOL */
	  if (l_Node) {
	    /* read in name */
            l_p=nwGetFirstQuote(l_p,l_word);
	    /* ok, let's search for something to match l_p */
            for (l_i=0;
	      (g_hwNodeSrv[l_i].nName)&&(stricmp(g_hwNodeSrv[l_i].nName,l_word));
	      l_i++);
	    if (g_hwNodeSrv[l_i].nName) {
	      /* we found our service! */
	      l_Node->nInitNode=g_hwNodeSrv[l_i].nInitNode;
              l_Node->nTickNode=g_hwNodeSrv[l_i].nTickNode;
	      l_Node->nKillNode=g_hwNodeSrv[l_i].nKillNode;
	      l_Node->nRcvPacket=g_hwNodeSrv[l_i].nRcvPacket;
	      l_Node->nDeliveryError=g_hwNodeSrv[l_i].nDeliveryError;
	    } else {
	      printf("ERROR: Invalid PROTOCOL specified: %s\n",
	        l_word);
	    }
	  } else {
	    printf("ERROR: PROTOCOL used without previous NODE\n");
	  }
	} else if (l_word[2]=='c') { /* PACKET_ALGORYTHM */
	  if (l_NodeCap) { 
	    /* read in name */
            l_p=nwGetFirstQuote(l_p,l_word);
	    /* ok, let's search for something to match l_p */
            for (l_i=0;
	      (g_hwNodeCapSrv[l_i].nName)&&
                (stricmp(g_hwNodeCapSrv[l_i].nName,l_word));
	      l_i++) {
	    }
	    if (g_hwNodeCapSrv[l_i].nName) {
	      /* we found our service! */
	      l_NodeCap->nInitNodeCap=g_hwNodeCapSrv[l_i].nInitNodeCap;
              l_NodeCap->nTickNodeCap=g_hwNodeCapSrv[l_i].nTickNodeCap;
	      l_NodeCap->nKillNodeCap=g_hwNodeCapSrv[l_i].nKillNodeCap;
	    } else {
	      printf("ERROR: Invalid PACKET_ALGORYTHM (NodeCap Srv): %s\n",
	        l_word);
	    }
	  } else {
	    printf("ERROR: PACKET_ALGORYTHM used without previous NODECAP\n");
	  }
	} else { /* PAUSETIME */
	  if (l_Net) {
            l_p=nwGetFirstWord(l_p,l_word);
            if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	      /* read in new Pause Time */
	      sscanf(l_word,"%lu",&(l_Net->nPauseTime));
	      l_Net->nPauseTime=NW_S(l_Net->nPauseTime);
	    }
	  } else {
	    printf("ERROR: PAUSETIME used without previous NETWORK\n");
	  }
        }
	break;
      case 's': /* SUBNETWORK, STOPTIME */
	if (l_word[1]=='u') { /* SUBNETWORK */
	  if (l_Net) {
	    /* First, create a subnetwork */
            l_SubNet=nwAllocSubNetwork(l_Net);
	    /* next, read in its name, if there is one */
            l_p=nwGetFirstQuote(l_p,l_SubNet->sName);
	    /* create a node for each NodeCap and link to NodeCap */
	    for (l_nc=l_Net->nNodeCap;l_nc;l_nc=l_nc->nNext) {
	      l_n=nwAllocNode(l_SubNet);
	      l_n->nNodeCap=l_nc;
	    }
	  } else {
	    printf("ERROR: SUBNETWORK used without previous NETWORK\n");
	  }
	} else { /* STOPTIME */
	  if (l_Net) {
            l_p=nwGetFirstWord(l_p,l_word);
            if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	      /* read in new Stop Time */
	      sscanf(l_word,"%lu",&(l_Net->nStopTime));
	      l_Net->nStopTime=NW_S(l_Net->nStopTime);
	    }
	  } else {
	    printf("ERROR: STOPTIME used without previous NETWORK\n");
	  }
	}
	break;
      case 'm': /* MEDIUM, MNAME */
        if (l_word[1]=='e') { /* MEDIUM */
	  if (l_SubNet) {
	    l_Medium=nwAllocMedium(l_SubNet);
	    /* check for ID specified */
            l_p=nwGetFirstWord(l_p,l_word);
            if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	      /* read in new ID */
	      sscanf(l_word,"%li",&(l_Medium->mID));
	    } 
	  } else {
	    printf("ERROR: MEDIUM used without previous SUBNETWORK\n");
	  }
	} else { /* MNAME */
          if (l_Medium) {
            l_p=nwGetFirstQuote(l_p,l_Medium->mName);
          } else {
	    printf("ERROR: MNAME used without previous MEDIUM\n");
          }
	}
	break;
      case 'b': /* BMULTIPLIER, BINCREMENTER, BAUD */
	if (l_word[1]=='m') { /* BMULTIPLIER  - use BAUD in your .net file instead */
	  if (l_Medium) {
	    /* read in baud multiplier */
            l_p=nwGetFirstWord(l_p,l_word);
            if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	      /* read in multiplier */
	      sscanf(l_word,"%li",&(l_Medium->mBaudMultiplier));
	      nwMultIncrToBaud(l_Medium);
	    }
	  } else { 
	    printf("ERROR: BMULT used without previous MEDIUM\n");
	  }
	} else if (l_word[1]=='a') { /* BAUD */
	  if (l_Medium) {
	    /* read in baud  */
            l_p=nwGetFirstWord(l_p,l_word);
            if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	      /* read in multiplier */
	      sscanf(l_word,"%f",&(l_Medium->mBaud));
	      nwBaudToMultIncr(l_Medium);
	    }
	  } else { 
	    printf("ERROR: BAUD used without previous MEDIUM\n");
	  }
	} else { /* BINCREMENTER - use BAUD in your .net file instead */
	  if (l_Medium) {
	    /* read in baud incrementer */
            l_p=nwGetFirstWord(l_p,l_word);
            if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	      /* read in incrementer */
	      sscanf(l_word,"%li",&(l_Medium->mBaudIncrementer));
	      nwMultIncrToBaud(l_Medium);
	    }
	  } else { 
	    printf("ERROR: BINC used without previous MEDIUM\n");
	  }
	}
	break;
      case 'i': /* INTERFACE */
	if (l_Node) {
	  /* first, read in our medium ID */
          l_p=nwGetFirstWord(l_p,l_word);
	  l_i=0;
          if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	    /* read in ID */
	    sscanf(l_word,"%li",&l_i);
	    /* now find that medium */
	    for (l_m=l_SubNet->sMedium;
              (l_m)&&(l_m->mID!=l_i);
              l_m=l_m->mNext);
            if (l_m) {
	      /* we have our node, l_Node, and our medium, l_m */
              nwAllocInterface(l_Node,l_m);
	    } else {
	      printf("ERROR: INTERFACE specified with invalid MEDIUM ID\n");
	    } 
	  } else {
	    printf("ERROR: INTERFACE specified with invalid ID\n");
	  }
	} else {
	  printf("ERROR: INTERFACE used without previous NODE\n");
	}
	break;
      case 't': /* TRANSPORT */
	if (l_Medium) {
	  /* read in name */
          l_p=nwGetFirstQuote(l_p,l_word);
	  /* ok, let's search for something to match l_p */
          for (l_i=0;
	    (g_hwMediumSrv[l_i].mName)&&
	      (stricmp(g_hwMediumSrv[l_i].mName,l_word));
	    l_i++);
	  if (g_hwMediumSrv[l_i].mName) {
	    /* we found our service! */
	    l_Medium->mInitMedium=g_hwMediumSrv[l_i].mInitMedium;
	    l_Medium->mTickMedium=g_hwMediumSrv[l_i].mTickMedium;
	    l_Medium->mKillMedium=g_hwMediumSrv[l_i].mKillMedium;
	    l_Medium->mKillInterface=g_hwMediumSrv[l_i].mKillInterface;
	    l_Medium->mSendPacket=g_hwMediumSrv[l_i].mSendPacket;
	    l_Medium->mMinHop=g_hwMediumSrv[l_i].mMinHop;
	  } else {
	    printf("ERROR: Invalid PACKET_ALGORYTHM (NodeCap Srv): %s\n",
	      l_word);
	  }
	} else {
	  printf("ERROR: TRANSPORT used without previous MEDIUM\n");
	}
	break;
      case 'e': /* EVENT */
        /* read in name */
	l_p=nwGetFirstWord(l_p,l_word);
	switch (l_word[0]) {
	  case 'm': /* EVENT MEDIUM - bring a medium up/down */
	    if (l_SubNet) {
	      /*  read in our medium ID */
              l_p=nwGetFirstWord(l_p,l_word);
  	      l_i=0;
              if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')) {
	        /* read in ID */
	        sscanf(l_word,"%li",&l_i);
	        /* now find that medium */
	        for (l_m=l_SubNet->sMedium;
                  (l_m)&&(l_m->mID!=l_i);
                  l_m=l_m->mNext);
                if (l_m) {
	          /* we have our medium in l_m */
		  /* let's read in our remaining parameters: Expire, Repeat, State */
                  l_p=nwGetFirstWord(l_p,l_word);
		  l_Exp=0;
	          sscanf(l_word,"%li",&l_Exp);
                  l_p=nwGetFirstWord(l_p,l_word);
		  l_Rep=0;
	          sscanf(l_word,"%li",&l_Rep);
                  l_p=nwGetFirstWord(l_p,l_word);
		  l_State=NWSTATE_INVALID;
	          sscanf(l_word,"%li",&l_State);
                  if (l_State) {
                    schScheduleEvent(NW_MS(l_Exp),NW_MS(l_Rep),&nwMediumEventStateOK,l_m);
		  } else {
                    schScheduleEvent(NW_MS(l_Exp),NW_MS(l_Rep),&nwMediumEventStateFail,l_m);
		  }
	        } else {
	          printf("ERROR: EVENT MEDIUM specified with invalid MEDIUM ID\n");
	        } 
	      } else {
	        printf("ERROR: EVENT type specified\n");
	      }
	    } else {
	      printf("ERROR: EVENT MEDIUM used without previous SUBNETWORK\n");
	    }
	    break;
	  case 'c': /* EVENT CONVERGENCE - turn on or off convergence reporting */
	    if (l_SubNet) {
              /* let's read in our remaining parameters: Expire, Repeat, State */
              l_p=nwGetFirstWord(l_p,l_word);
	      l_Exp=0;
	      sscanf(l_word,"%li",&l_Exp);
              l_p=nwGetFirstWord(l_p,l_word);
	      l_Rep=0;
	      sscanf(l_word,"%li",&l_Rep);
              l_p=nwGetFirstWord(l_p,l_word);
	      l_State=0;
	      sscanf(l_word,"%li",&l_State);
              if (l_State) {
                schScheduleEvent(NW_MS(l_Exp),NW_MS(l_Rep),&nwSubNetConvergenceReportOn,l_SubNet);
	      } else {
                schScheduleEvent(NW_MS(l_Exp),NW_MS(l_Rep),&nwSubNetConvergenceReportOff,l_SubNet);
	      }
	    } else {
	      printf("ERROR: EVENT CONVERGENCE used without previous SUBNETWORK\n");
	    }
            break;
	  default:
	    printf("ERROR: Unknown EVENT command %s in file %s.\n",l_word,l_File);
	    break;
	}
	break;
      case 'd': /* DELAY */
	if (l_Medium) {
	  /* read in propagation delay */
          l_p=nwGetFirstWord(l_p,l_word);
          if ( ((l_word[0]>='0')&&(l_word[0]<='9'))||(l_word[0]=='-')||(l_word[0]=='.')) {
	    /* read in multiplier */
	    sscanf(l_word,"%f",&l_Medium->mPropagationDelay);
	  }
	} else { 
	  printf("ERROR: DELAY used without previous MEDIUM\n");
	}
	break;
      default:
	printf("ERROR: Unknown command %s in file %s.\n",l_word,l_File);
	break;
    } 
  }
  fclose(l_infile);
  return l_Net;
}

