/* Packet Source Module */

/* Options */
/* #define PSNOISY*/

/* #ifdef PSNOISY */
#include<stdio.h>
/* #endif */
#include<stdlib.h>
#include<math.h>
#include"memmgr.h"
#include"network.h"
#include"pksource.h"


/*********************************/
/***** Constant Stream Suite *****/
/*********************************/
/*** Parameter 0 : destination address 
 *** Parameter 1 : Period (in ms) - packets send every Param1 ms 
 ***/
void psCSInitNodeCap(NWNODECAP *p_NodeCap) {
  PSCONSTR *l_c;

#ifdef PSNOISY 
  printf("ConStr: Init Nodecap %li\n",p_NodeCap->nID);
#endif

  l_c=p_NodeCap->nGeneral=mmuAlloc(sizeof(PSCONSTR));
  l_c->cDestID=p_NodeCap->nParam[0];
  /* l_c->cTickDivisor=200000+rand()%300000;*/
  /* l_c->cTickDivisor=NW_MS(400); */
  l_c->cTickDivisor=NW_MS(p_NodeCap->nParam[1]);
  l_c->cTickMod=rand()%l_c->cTickDivisor; 
}

void psCSTickNodeCap(NWNODECAP *p_NodeCap) {
  PSCONSTR *l_c;
  l_c=p_NodeCap->nGeneral;

/*
#ifdef PSNOISY 
  printf("ConStr: Tick Nodecap %li\n",p_NodeCap->nID);
#endif
*/

  if (l_c->cDestID<0) {
    NWNODECAP *l_NodeCap;
    long l_i,l_j,l_k=0;
    /* we need to find a destination */
    for(l_i=0,l_NodeCap=p_NodeCap->nNet->nNodeCap;
      l_NodeCap;
      l_NodeCap=l_NodeCap->nNext,l_i++) {
      if (l_NodeCap==p_NodeCap) 
        l_k=l_i;
    }
    l_j=l_k;
    while(l_j==l_k)
      l_j=rand()%l_i;

    for(l_i=0,l_NodeCap=p_NodeCap->nNet->nNodeCap;
      l_NodeCap&&(l_i!=l_j);
      l_NodeCap=l_NodeCap->nNext,l_i++);
    l_c->cDestID=l_NodeCap->nID;
  }

  if ((p_NodeCap->nNet->nTick%l_c->cTickDivisor)==l_c->cTickMod) {
    /* source packet */
#ifdef PSNOISY
    printf("ConStr: Sourcing Packet from %li to %li!\n",
      p_NodeCap->nID,l_c->cDestID);
#endif
    /*nwSourcePacket(p_NodeCap,rand()%1000,l_c->cDestID);*/
    /* nwSourcePacket(p_NodeCap,rand()%100,l_c->cDestID);*/
    nwSourcePacket(p_NodeCap,32,l_c->cDestID);
  }
}

void psCSKillNodeCap(NWNODECAP *p_NodeCap) {
#ifdef PSNOISY 
  printf("ConStr: Kill NodeCap %li!\n",p_NodeCap->nID);
#endif
  mmuFree(p_NodeCap->nGeneral);
}

/*********************************/
/***** Event (random) Suite  *****/
/*********************************/
/*** Parameter 0 : destination address 
 *** Parameter 1 : Base Period (in ms)  
 *** Parameter 2 : Variance (+- this amount)
 *** Period = Base (P1) - P2 + rand()%P2 + rand()%P2 
 ***   Two rand()'s give a non-linear randomness - hedging 
 ***   towards gaussian around the middle.
 ***/
void psEventInitNodeCap(NWNODECAP *p_NodeCap) {
  PSCONSTR *l_c;

#ifdef PSNOISY 
  printf("EventSrc: Init Nodecap %li\n",p_NodeCap->nID);
#endif

  l_c=p_NodeCap->nGeneral=mmuAlloc(sizeof(PSCONSTR));
  l_c->cDestID=p_NodeCap->nParam[0];
  l_c->cTickBase=NW_MS(p_NodeCap->nParam[1]);
  l_c->cTickVariance=NW_MS(p_NodeCap->nParam[2]);
  l_c->cTickBase-=l_c->cTickVariance;
  l_c->cTicksLeft=p_NodeCap->nNet->nTick+l_c->cTickBase+rand()%l_c->cTickVariance+rand()%l_c->cTickVariance;

}

void psEventTickNodeCap(NWNODECAP *p_NodeCap) {
  PSCONSTR *l_c;
  l_c=p_NodeCap->nGeneral;

/*
#ifdef PSNOISY 
  printf("EventSrc: Tick Nodecap %li\n",p_NodeCap->nID);
#endif
*/

  if (l_c->cDestID<0) {
    NWNODECAP *l_NodeCap;
    long l_i,l_j,l_k=0;
    /* we need to find a destination */
    for(l_i=0,l_NodeCap=p_NodeCap->nNet->nNodeCap;
      l_NodeCap;
      l_NodeCap=l_NodeCap->nNext,l_i++) {
      if (l_NodeCap==p_NodeCap) 
        l_k=l_i;
    }
    l_j=l_k;
    while(l_j==l_k)
      l_j=rand()%l_i;

    for(l_i=0,l_NodeCap=p_NodeCap->nNet->nNodeCap;
      l_NodeCap&&(l_i!=l_j);
      l_NodeCap=l_NodeCap->nNext,l_i++);
    l_c->cDestID=l_NodeCap->nID;
  }

  if (p_NodeCap->nNet->nTick>=l_c->cTicksLeft) {
    /* source packet */
#ifdef PSNOISY
    printf("EventSrc: Sourcing Packet from %li to %li!\n",
      p_NodeCap->nID,l_c->cDestID);
#endif
    /*nwSourcePacket(p_NodeCap,rand()%1000,l_c->cDestID);*/
    /* nwSourcePacket(p_NodeCap,rand()%100,l_c->cDestID);*/
    nwSourcePacket(p_NodeCap,32,l_c->cDestID);
    l_c->cTicksLeft=p_NodeCap->nNet->nTick+l_c->cTickBase+rand()%l_c->cTickVariance+rand()%l_c->cTickVariance;
  }
}

void psEventKillNodeCap(NWNODECAP *p_NodeCap) {
#ifdef PSNOISY 
  printf("EventSrc: Kill NodeCap %li!\n",p_NodeCap->nID);
#endif
  mmuFree(p_NodeCap->nGeneral);
}


/**********************************/
/***** Multi-Connection Suite *****/
/**********************************/

#define PSMULTI_CONNECTION_HOLDOFF(x) \
 ( (rand()%x)+(rand()%x) )/2;

void psMultiInitNodeCap(NWNODECAP *p_NodeCap) {
  PSMULTICON *l_c;
  long l_i;

#ifdef PSNOISY 
  printf("MultiCon: Init Nodecap %li\n",p_NodeCap->nID);
#endif

  l_c=p_NodeCap->nGeneral=mmuAlloc(sizeof(PSMULTICON));
  for (l_i=0;l_i<PS_MULTICON_NUMBER;l_i++) {
    l_c->cDestID[l_i]=NWADDR_INVALID;
    l_c->cHoldoff[l_i]=PSMULTI_CONNECTION_HOLDOFF(NW_S(15));
    l_c->cPause=p_NodeCap->nNet->nPauseTime;
    l_c->cMoveTicks=0L; 
  }
  if (p_NodeCap->nX<0)
    p_NodeCap->nX=rand()%(9*NWMETER);
  if (p_NodeCap->nY<0)
    p_NodeCap->nY=rand()%(9*NWMETER);
}

void psMultiTickNodeCap(NWNODECAP *p_NodeCap) {
  PSMULTICON *l_c;
  NWNODECAP *l_NodeCap;
  long l_i,l_j,l_k=0,l_m;
  double l_dX,l_dY,l_dM;
  NWPACKET *l_Packet;

  l_c=p_NodeCap->nGeneral;

/*
#ifdef PSNOISY 
  printf("MultiCon: Tick Nodecap %li\n",p_NodeCap->nID);
#endif
*/

  /* let's see if we need to move */
  if (l_c->cMoveTicks) {
    l_c->cMoveTicks--;
    p_NodeCap->nX+=l_c->cDX;
    p_NodeCap->nY+=l_c->cDY;
  } else if (l_c->cPause) {
    l_c->cPause--;
  } else {
#ifdef PSNOISY 
    printf("MultiCon: Node %li (%li,%li) picking new move/pause parameters\n",
      p_NodeCap->nID,p_NodeCap->nX,p_NodeCap->nY);
#endif
    /* Pick a new destination & pause time for this sucker */
    l_c->cPause=p_NodeCap->nNet->nPauseTime;
    l_i=rand()%(9*NWMETER); /* new X */
    l_j=rand()%(9*NWMETER); /* new Y */
    l_m=rand()%(4*(NWMETER/NWTICK))+(3*(NWMETER/NWTICK));
    
    /* l_i=new X, l_j=new Y, l_m=movement rate */
    l_dX=(l_i-p_NodeCap->nX);
    l_dY=(l_j-p_NodeCap->nY);
    l_dM=l_m;
    l_c->cMoveTicks=(sqrt((l_dX*l_dX)+(l_dY*l_dY))/l_dM);

    /* ok, I think we can dispense with the floats now */
    if (l_c->cMoveTicks) {
      l_c->cDX=(l_i-p_NodeCap->nX)/l_c->cMoveTicks;
      l_c->cDY=(l_j-p_NodeCap->nY)/l_c->cMoveTicks;
    }
  }


  /* check existing connections -- see if we should source a packet */
  for (l_i=0;l_i<PS_MULTICON_NUMBER;l_i++) {
    if (l_c->cDestID[l_i]!=NWADDR_INVALID) {
      if ((p_NodeCap->nNet->nTick%l_c->cTickDivisor[l_i])==l_c->cTickMod[l_i]) {
        /* source packet */
        l_Packet=nwSourcePacket(p_NodeCap,
          (rand()%100<70)?1000:32,
          l_c->cDestID[l_i]);
#ifdef PSNOISY
        printf("MultiCon: Sourcing Packet %li from %li to %li!\n",
          l_Packet->pID,p_NodeCap->nID,l_c->cDestID[l_i]);
#endif
        l_c->cPkCount[l_i]--;
        if (l_c->cPkCount[l_i]<=0) { /* conversation is done */
#ifdef PSNOISY 
          printf("MultiCon: Node %li closing connection to %li\n",
            p_NodeCap->nID,l_c->cDestID[l_i]);
#endif
          l_c->cDestID[l_i]=NWADDR_INVALID;
          l_j=PSMULTI_CONNECTION_HOLDOFF(NW_S(15));
          for (l_k=0;l_k<PS_MULTICON_NUMBER;l_k++) {
            l_c->cHoldoff[l_k]=l_j;
          }
        }
      }
    }
  }

  l_j=0;
  /* now see if we should start a new connection */
  for (l_i=0;l_i<PS_MULTICON_NUMBER;l_i++) {
    if (l_c->cHoldoff[l_i])
      l_c->cHoldoff[l_i]--;
    if (l_c->cDestID[l_i]==NWADDR_INVALID) 
      l_j=1;
  }

  if (l_j) {
    /* we should check to see if we need to make a new connection */
    l_m=-1;
    for (l_i=0;l_i<PS_MULTICON_NUMBER;l_i++) {
      if ((l_c->cDestID[l_i]==NWADDR_INVALID) &&
          (l_c->cHoldoff[l_i]==0)) {
        l_m=l_i; 
      }
    }
    if (l_m>=0) {
      /* yep, let's find a destination */
      for(l_i=0,l_NodeCap=p_NodeCap->nNet->nNodeCap;
        l_NodeCap;
        l_NodeCap=l_NodeCap->nNext,l_i++) {
        if (l_NodeCap==p_NodeCap) 
          l_k=l_i;
      }
      l_j=l_k;
      while(l_j==l_k)
        l_j=rand()%l_i;

      for(l_i=0,l_NodeCap=p_NodeCap->nNet->nNodeCap;
        l_NodeCap&&(l_i!=l_j);
        l_NodeCap=l_NodeCap->nNext,l_i++);
      l_c->cDestID[l_m]=l_NodeCap->nID;

#ifdef PSNOISY 
      printf("MultiCon: Node %li forming connection to %li\n",
        p_NodeCap->nID,l_NodeCap->nID);
#endif

      /* we also need to select some packet sourcing parameters too */
      l_c->cTickDivisor[l_m]=NW_MS(200)+rand()%NW_MS(300);
      l_c->cTickMod[l_m]=rand()%(l_c->cTickDivisor[l_m]);
      l_c->cPkCount[l_m]=rand()%1000+500;

      /* now holdoff other connections from forming */
      l_j=PSMULTI_CONNECTION_HOLDOFF(NW_S(15)) \
      for (l_i=0;l_i<PS_MULTICON_NUMBER;l_i++) {
        l_c->cHoldoff[l_i]=l_j;
      }
    }
  }
}

void psMultiKillNodeCap(NWNODECAP *p_NodeCap) {
#ifdef PSNOISY 
  printf("MultiCon: Kill NodeCap %li!\n",p_NodeCap->nID);
#endif
  mmuFree(p_NodeCap->nGeneral);
}

