/* WIRE.H 
 * Two-letter code: wm
 */

/* Statistics info for mStats */
#define WM_S_PACKETS_SOURCE      0   /* # packets sent to us for transmission */
#define WM_S_PACKETS_SENT        1   /* # transmission attempts */
#define WM_S_PACKETS_LOSTRANDOM  2   /* # corrupted transmissions - random */
#define WM_S_PACKETS_LOSTLINK    3   /* # lost due to link being down or going down */
#define WM_S_PACKETS_TRANSOK     4   /* # packets transmitted OK */
#define WM_S_RETRANSMISSIONS     5   /* # retransmissions */
#define WM_S_COUNT               6   /* # statistics collected */

/* Structures */
typedef struct WMINTERFACEDATA {
  NWPACKET *iPacketBuffer;  /* holds one packet -- waiting or being sent */
  long iNodePriority;       /* increases each time node has something to send but can't */
  long iNodeSendHoldoff;    /* holdoff before sending */
  long iRetransmit;         /* # retransmissions left */
} WMINTERFACEDATA;

typedef struct WMMEDIUMDATA {
  NWINTERFACE *mSourceInt; /* points to the sender, NULL if none */
  long mTicksToSend; /* if sending, # ticks (not bytes) required total to send */
  long mTicksSent; /* if sending, # ticks (not bytes) sent so far. */
  long mSendHoldoff; /* delay causing inability for anyone to transmit. This could be 
                      * used to simulate the time consumed by a transmission collision */
  long mCorrupted;   /* This flag gets set if the packet gets corrupted. It still gets "sent" 
                      * for timing etc., but as soon as it becomes corrupted, a NAM "drop" 
		      * signal is sent, and the packet is never delivered. */
  /* Statistics */
  long mStats[WM_S_COUNT];
} WMMEDIUMDATA;

/* Function prototypes */
void wmWireInitMedium(NWMEDIUM *p_Medium);
void wmWireTickMedium(NWMEDIUM *p_Medium);
void wmWireKillMedium(NWMEDIUM *p_Medium);
void wmWireKillInterface(NWINTERFACE *p_Interface);
unsigned long wmWireSendPacket(NWINTERFACE *p_Interface, NWPACKET *p_Packet, long p_DestinationID);
void wmWireStatCount(NWMEDIUM *p_Medium,int p_Count);
void wmWireStatPrint(NWSUBNETWORK *p_SubNet,int p_Flag);

